from mage_ai.data_cleaner import column_type_detector
from mage_ai.data_cleaner.analysis.calculator import AnalysisCalculator
from mage_ai.data_cleaner.pipelines.base import BasePipeline
from mage_ai.data_cleaner.shared.hash import merge_dict
from mage_ai.data_cleaner.shared.logger import timer
from mage_ai.data_cleaner.shared.utils import clean_dataframe
from mage_ai.data_cleaner.statistics.calculator import StatisticsCalculator


def analyze(df):
    cleaner = DataCleaner()
    return cleaner.analyze(df)


def clean(df, transform=True):
    cleaner = DataCleaner()
    return cleaner.clean(df, transform=transform)


class DataCleaner():
    def analyze(self, df):
        """ Analyze a dataframe
        1. Detect column types
        2. Calculate statisitics
        3. Calculate analysis
        """
        with timer('data_cleaner.infer_column_types'):
            column_types = column_type_detector.infer_column_types(df)
        with timer('data_cleaner.clean_series'):
            df = clean_dataframe(df, column_types, dropna=False)
        with timer('data_cleaner.calculate_statistics'):
            statistics = StatisticsCalculator(column_types).process(df, is_clean=True)
        with timer('data_cleaner.calculate_insights'):
            analysis = AnalysisCalculator(df, column_types, statistics).process(df, is_clean=True)
        return dict(
            insights=analysis,
            column_types=column_types,
            statistics=statistics,
        )

    def clean(self, df, transform=True):
        df_stats = self.analyze(df)
        df = clean_dataframe(df, df_stats['column_types'])
        pipeline = BasePipeline()
        if df_stats['statistics']['is_timeseries']:
            df = df.sort_values(by=df_stats['statistics']['timeseries_index'], axis=0)
        # TODO: Pass in both cleaned and uncleaned versions of dataset
        with timer('data_cleaner.create_suggested_actions'):
            suggested_actions = pipeline.create_actions(
                df,
                df_stats['column_types'],
                df_stats['statistics'],
            )
        if transform:
            with timer('data_cleaner.transform_data'):
                df_transformed = pipeline.transform(df, auto=True)
        else:
            df_transformed = df
        return merge_dict(df_stats, dict(
            df=df_transformed,
            suggestions=suggested_actions,
            pipeline=pipeline,
        ))



from mage_ai.data_cleaner.shared.multi import run_parallel_multiple_args
import numpy as np
import pandas as pd
import re

DATETIME_MATCHES_THRESHOLD = 0.5
MAXIMUM_WORD_LENGTH_FOR_CATEGORY_FEATURES = 40
MULTITHREAD_MAX_NUM_ENTRIES = 50000
NUMBER_TYPE_MATCHES_THRESHOLD = 0.8
STRING_TYPE_MATCHES_THRESHOLD = 0.3

CATEGORY = 'category'
CATEGORY_HIGH_CARDINALITY = 'category_high_cardinality'
DATETIME = 'datetime'
EMAIL = 'email'
NUMBER = 'number'
NUMBER_WITH_DECIMALS = 'number_with_decimals'
PHONE_NUMBER = 'phone_number'
TEXT = 'text'
TRUE_OR_FALSE = 'true_or_false'
ZIP_CODE = 'zip_code'

CATEGORICAL_TYPES = frozenset([CATEGORY, CATEGORY_HIGH_CARDINALITY, TRUE_OR_FALSE])
NUMBER_TYPES = frozenset([NUMBER, NUMBER_WITH_DECIMALS])
STRING_TYPES = frozenset([EMAIL, PHONE_NUMBER, TEXT, ZIP_CODE])

COLUMN_TYPES = frozenset(
    [
        CATEGORY,
        CATEGORY_HIGH_CARDINALITY,
        DATETIME,
        EMAIL,
        NUMBER,
        NUMBER_WITH_DECIMALS,
        PHONE_NUMBER,
        TEXT,
        TRUE_OR_FALSE,
        ZIP_CODE,
    ]
)

REGEX_DATETIME_PATTERN = r'^\d{2,4}-\d{1,2}-\d{1,2}$|^\d{2,4}-\d{1,2}-\d{1,2}[Tt ]{1}\d{1,2}:\d{1,2}[:]{0,1}\d{1,2}[\.]{0,1}\d*|^\d{4}-\d{2}-\d{2}T\d{2}:\d{2}:\d{2}\.\d{3}Z$|^\d{1,4}[-\/]{1}\d{1,2}[-\/]{1}\d{1,2}$|^\d{1,2}[-\/]{1}\d{1,2}[-\/]{1}\d{1,4}$|^\d{1,2}[-\/]\d{1,2}[-\/]\d{2,4}[Tt ]{1}\d{1,2}:\d{1,2}[:]{0,1}\d{1,2}[\.]{0,1}\d*|(Jan(?:uary)?|Feb(?:ruary)?|Mar(?:ch)?|Apr(?:il)?|May|Jun(?:e)?|Jul(?:y)?|Aug(?:ust)?|Sep(?:tember)?|Oct(?:ober)?|Nov(?:ember)?|Dec(?:ember)?)\s+(\d{1,2})[\s,]+(\d{2,4})'
REGEX_DATETIME = re.compile(REGEX_DATETIME_PATTERN)
REGEX_EMAIL_PATTERN = r'^[a-zA-Z0-9_.+-]+@[a-zA-Z0-9-]+\.[a-zA-Z0-9-.]+$'
REGEX_EMAIL = re.compile(REGEX_EMAIL_PATTERN)
REGEX_INTEGER_PATTERN = r'^\-{0,1}\s*(?:(?:[$€¥₹£]|Rs|CAD){0,1}\s*(?:[0-9]+(?:,[0-9]+)*|[0-9]+){0,1}|(?:[0-9]+(?:,[0-9]+)*|[0-9]+){0,1}\s*(?:[元€$]|CAD){0,1})$'
REGEX_INTEGER = re.compile(REGEX_INTEGER_PATTERN)
REGEX_FLOAT_NEW_SYM = re.compile(r'[\.\%]')
REGEX_NUMBER_PATTERN = r'^\-{0,1}\s*(?:(?:[$€¥₹£]|Rs|CAD){0,1}\s*(?:[0-9]+(?:,[0-9]+)*|[0-9]+){0,1}(?:\.[0-9]*){0,1}|(?:[0-9]+(?:,[0-9]+)*|[0-9]+){0,1}(?:\.[0-9]*){0,1}\s*(?:[元€$]|CAD){0,1})\s*\%{0,1}$'
REGEX_NUMBER = re.compile(REGEX_NUMBER_PATTERN)
REGEX_PHONE_NUMBER_PATTERN = (
    r'^\s*(?:\+?(\d{1,3}))?[-. (]*(\d{3})[-. )]*(\d{3})[-. ]*(\d{4})(?: *x(\d+))?\s*$'
)
REGEX_PHONE_NUMBER = re.compile(REGEX_PHONE_NUMBER_PATTERN)
REGEX_ZIP_CODE_PATTERN = r'^\d{3,5}(?:[-\s]\d{4})?$'
REGEX_ZIP_CODE = re.compile(REGEX_ZIP_CODE_PATTERN)

RESERVED_PHONE_NUMBER_WORDS = frozenset(['phone', 'landline'])
RESERVED_ZIP_CODE_WORDS = frozenset(['zip', 'postal', 'zipcode', 'postcode'])


def str_in_set(string, string_set):
    return any(entry in string for entry in string_set)


def get_mismatched_rows(series, column_type):
    if len(series) == 0:
        return []
    mismatched_rows = []
    if column_type == EMAIL:
        mismatched_rows = series[~series.str.match(REGEX_EMAIL)].tolist()
    elif column_type == PHONE_NUMBER:
        mismatched_rows = series[~series.str.match(REGEX_PHONE_NUMBER)].tolist()
    elif column_type == ZIP_CODE:
        str_series = series.astype(str)
        mismatched_rows = series[~str_series.str.match(REGEX_ZIP_CODE)].tolist()
    return mismatched_rows


def infer_number_type(series, column_name, dtype):
    clean_series = series.dropna()
    length = len(clean_series)
    if length == 0:
        mdtype = NUMBER_WITH_DECIMALS
    else:
        correct_phone_nums = (
            (clean_series >= 1e9) & (clean_series < 1e12) & (np.floor(clean_series) == clean_series)
        ).sum()
        if (
            correct_phone_nums / length >= NUMBER_TYPE_MATCHES_THRESHOLD
            and 'phone' in column_name.lower()
        ):
            mdtype = PHONE_NUMBER
        else:
            if np.issubdtype(dtype, np.integer):
                if (
                    clean_series.min() >= 100
                    and clean_series.max() <= 99999
                    and str_in_set(column_name.lower(), RESERVED_ZIP_CODE_WORDS)
                ):
                    mdtype = ZIP_CODE
                else:
                    mdtype = NUMBER
            elif np.issubdtype(dtype, np.floating):
                mdtype = NUMBER_WITH_DECIMALS
    return mdtype


def infer_column_type(series, column_name, dtype, kwargs):
    mdtype = None
    if 'datetime64' in str(dtype):
        mdtype = DATETIME
    elif dtype == 'object':
        mdtype = infer_object_type(series, column_name, kwargs)
    elif dtype == 'bool':
        mdtype = TRUE_OR_FALSE
    elif np.issubdtype(dtype, np.floating) or np.issubdtype(dtype, np.integer):
        mdtype = infer_number_type(series, column_name, dtype)
    if mdtype in NUMBER_TYPES and series.nunique(dropna=False) == 2:
        mdtype = TRUE_OR_FALSE

    return mdtype


def infer_object_type(series, column_name, kwargs):
    clean_series = series.apply(lambda x: x.strip(' \'\"') if type(x) is str else x)
    clean_series = clean_series.map(lambda x: x if (not isinstance(x, str) or x != '') else np.nan)
    clean_series = clean_series.dropna().astype(str)

    series_nunique = series.nunique(dropna=False)
    clean_series_nunique = clean_series.nunique()

    exact_dtype = type(clean_series.iloc[0]) if clean_series.count() else None
    if exact_dtype is list:
        mdtype = TEXT
    elif np.issubdtype(exact_dtype, np.bool_):
        if clean_series_nunique <= 2:
            mdtype = TRUE_OR_FALSE
        else:
            mdtype = CATEGORY
    elif clean_series_nunique <= 2:
        mdtype = TRUE_OR_FALSE
    else:
        length = len(clean_series)
        if all(clean_series.str.match(REGEX_NUMBER)):
            if clean_series.str.contains(REGEX_FLOAT_NEW_SYM).sum():
                mdtype = NUMBER_WITH_DECIMALS
            else:
                lowercase_column_name = column_name.lower()
                correct_phone_nums = clean_series.str.match(REGEX_PHONE_NUMBER).sum()
                correct_zip_codes = clean_series.str.match(REGEX_ZIP_CODE).sum()
                if correct_phone_nums / length >= NUMBER_TYPE_MATCHES_THRESHOLD and str_in_set(
                    lowercase_column_name, RESERVED_PHONE_NUMBER_WORDS
                ):
                    mdtype = PHONE_NUMBER
                elif correct_zip_codes / length >= NUMBER_TYPE_MATCHES_THRESHOLD and str_in_set(
                    lowercase_column_name, RESERVED_ZIP_CODE_WORDS
                ):
                    mdtype = ZIP_CODE
                else:
                    mdtype = NUMBER
        else:
            matches = clean_series.str.match(REGEX_DATETIME).sum()
            if matches / length >= DATETIME_MATCHES_THRESHOLD:
                mdtype = DATETIME
            else:
                correct_emails = clean_series.str.match(REGEX_EMAIL).sum()
                correct_phone_nums = clean_series.str.match(REGEX_PHONE_NUMBER).sum()
                correct_zip_codes = clean_series.str.match(REGEX_ZIP_CODE).sum()
                lowercase_column_name = column_name.lower()
                if correct_emails / length >= STRING_TYPE_MATCHES_THRESHOLD:
                    mdtype = EMAIL
                elif correct_phone_nums / length >= STRING_TYPE_MATCHES_THRESHOLD and str_in_set(
                    lowercase_column_name, RESERVED_PHONE_NUMBER_WORDS
                ):
                    mdtype = PHONE_NUMBER
                elif correct_zip_codes / length >= STRING_TYPE_MATCHES_THRESHOLD and str_in_set(
                    lowercase_column_name, RESERVED_ZIP_CODE_WORDS
                ):
                    mdtype = ZIP_CODE
                elif series_nunique == 2:
                    mdtype = TRUE_OR_FALSE
                else:
                    if type(exact_dtype) is list:
                        mdtype = TEXT
                    elif clean_series_nunique / length >= 0.8:
                        mdtype = TEXT
                    else:
                        word_count = clean_series.map(lambda x: len(str(x).split(' '))).max()
                        if word_count > MAXIMUM_WORD_LENGTH_FOR_CATEGORY_FEATURES:
                            mdtype = TEXT
                        else:
                            if clean_series_nunique <= kwargs.get(
                                'category_cardinality_threshold', 255
                            ):
                                mdtype = CATEGORY
                            else:
                                mdtype = CATEGORY_HIGH_CARDINALITY
    return mdtype


def infer_column_types(df, **kwargs):
    columns = [df[col] for col in df.columns]
    kwarg_list = [kwargs] * len(df.columns)
    ctypes = {}
    num_entries = len(df)
    if num_entries > MULTITHREAD_MAX_NUM_ENTRIES:
        types = run_parallel_multiple_args(
            infer_column_type, columns, df.columns, df.dtypes, kwarg_list
        )
    else:
        types = map(infer_column_type, columns, df.columns, df.dtypes, kwarg_list)
    for col, dtype in zip(df.columns, types):
        ctypes[col] = dtype
    return ctypes



from faker import Faker
from mage_ai.data_cleaner.column_type_detector import (
    CATEGORY,
    DATETIME,
    EMAIL,
    MAXIMUM_WORD_LENGTH_FOR_CATEGORY_FEATURES,
    NUMBER,
    NUMBER_WITH_DECIMALS,
    PHONE_NUMBER,
    TEXT,
    TRUE_OR_FALSE,
    ZIP_CODE,
    get_mismatched_rows,
    infer_column_types,
    REGEX_NUMBER,
)
from mage_ai.tests.base_test import TestCase
import pandas as pd
import numpy as np


class ColumnTypeDetectorTests(TestCase):
    def setUp(self):
        Faker.seed(42)
        self.fake = Faker()
        return super().setUp()

    def test_get_mismatched_rows(self):
        df = pd.DataFrame(
            [
                [1, 'abc@xyz.com', '32132'],
                [2, 'abc2@xyz.com', '12345'],
                [3, 'test', '1234'],
                [4, 'abc@test.net', 'abcde'],
                [5, 'abc12345@', '54321'],
                [6, 'abcdef@123.com', '56789'],
            ],
            columns=['id', 'email', 'zip_code'],
        )
        rows1 = get_mismatched_rows(df['id'], 'number')
        rows2 = get_mismatched_rows(df['email'], 'email')
        rows3 = get_mismatched_rows(df['zip_code'], 'zip_code')
        self.assertEqual(rows1, [])
        self.assertEqual(rows2, ['test', 'abc12345@'])
        self.assertEqual(rows3, ['abcde'])

    def test_number_pattern_matching(self):
        good_number_patterns = [
            '3123452',
            '23.234232',
            '-23,245,234.93',
            '.934',
            '-0.000002',
            '$ 0.043',
            '-€ 23,456,789.00092',
            '¥ 21345634',
            '- Rs   345.23423523',
            'CAD 2,345,234,345.23423',
            '234.23423423 $',
            '-8907987 CAD',
            '3239458734 元',
            '234,2342,2343.823034234%',
            '234.3433   %',
        ]
        bad_number_patterns = [
            'this is not a number',
            'NOTCURR 23423933',
            '0.000.00.000.000',
            '1-111-111-111',
            ',2343234203',
            'ksjh&A(*2jnkj&#$#$',
        ]

        for good in good_number_patterns:
            print(good)
            match = REGEX_NUMBER.match(good)
            self.assertIsNotNone(match)

        for bad in bad_number_patterns:
            match = REGEX_NUMBER.match(bad)
            self.assertIsNone(match)

    def test_infer_column_types(self):
        columns = [
            'true_or_false',
            'number_with_decimals',
            'category',
            'datetime',
            'text',
            'number',
            'number_with_dollars',
            'number_with_percentage',
            'zip_code',
            'zip_code_with_3_numbers',
            'invalid_zip_code',
            'email',
            'phone_number',
            'datetime_abnormal',
            'name',
        ]
        table = [
            [
                '1',
                '3',
                'male',
                '2020-1-1',
                '1.0',
                '1',
                3,
                '30%',
                '10128-1213',
                123,
                123,
                'fire@mage.ai',
                '123-456-7890',
                'May 4, 2021, 6:35 PM',
                self.fake.name(),
            ],
            [
                '1',
                '12.0',
                'female',
                '2020-07-13',
                ' '.join(['t' for i in range(MAXIMUM_WORD_LENGTH_FOR_CATEGORY_FEATURES + 1)]),
                '2',
                '€4',
                '12.32%',
                12345,
                1234,
                1234,
                'mage@fire.com',
                '(123) 456-7890',
                'Feb 17, 2021, 2:57 PM',
                self.fake.name(),
            ],
            [
                '1',
                '0',
                'machine',
                '2020-06-25 01:02',
                ' '.join(['t' for i in range(MAXIMUM_WORD_LENGTH_FOR_CATEGORY_FEATURES + 1)]),
                '3',
                '¥5,000',
                '50%',
                '12345',
                12345,
                12345,
                'fire_mage@mage.com',
                '1234567890',
                'Feb 18, 2021, 2:57 PM',
                self.fake.name(),
            ],
            [
                0,
                '40.7',
                'mutant',
                '2020-12-25 01:02:03',
                ' '.join(['t' for i in range(MAXIMUM_WORD_LENGTH_FOR_CATEGORY_FEATURES + 1)]),
                '4',
                'Rs 5,000.01',
                '20%',
                '12345',
                12345,
                123456,
                'mage-fire@mage.ai',
                '1234567',
                'Feb 19, 2021, 2:57 PM',
                self.fake.name(),
            ],
            [
                0,
                '40.7',
                'alien',
                '2020-12-25T01:02:03.000Z',
                ' '.join(['t' for i in range(MAXIMUM_WORD_LENGTH_FOR_CATEGORY_FEATURES + 1)]),
                '4',
                '-10128,121.3123元',
                '18%',
                '01234',
                12345,
                12,
                'fire+1@mage.ai',
                '(123)456-7890',
                'Feb 20, 2021, 2:57 PM',
                self.fake.name(),
            ],
        ]
        date_formats = [
            '01/1/2019',
            '1/1/2019',
            '1/21/2019',
            '11/1/2019',
            '2020/01/1',
            '2020/1/01',
            '2020/1/1',
            'Pending',
        ]

        for date_format in date_formats:
            table.append(
                [
                    0,
                    '40.7',
                    'mutant',
                    date_format,
                    self.fake.text(),
                    4,
                    '$5,000.01',
                    '15.32%',
                    '01234',
                    12345,
                    12,
                    'fire+1@mage.ai',
                    '(123)456-7890',
                    'Feb 18, 2021, 2:57 PM',
                    self.fake.name(),
                ]
            )

        df = pd.DataFrame(table, columns=columns)
        column_types = infer_column_types(df)

        self.assertEqual(
            column_types,
            {
                'true_or_false': TRUE_OR_FALSE,
                'number_with_decimals': NUMBER_WITH_DECIMALS,
                'category': CATEGORY,
                'datetime': DATETIME,
                'text': TEXT,
                'number': NUMBER,
                'number_with_dollars': NUMBER_WITH_DECIMALS,
                'number_with_percentage': NUMBER_WITH_DECIMALS,
                'zip_code': ZIP_CODE,
                'zip_code_with_3_numbers': ZIP_CODE,
                'invalid_zip_code': NUMBER,
                'email': EMAIL,
                'phone_number': PHONE_NUMBER,
                'datetime_abnormal': DATETIME,
                'name': TEXT,
            },
        )

    def test_phone_number_recognition(self):
        df = pd.DataFrame(
            dict(
                phone_num=[self.fake.phone_number() for _ in range(6)],
                str_int_phone_nums=[
                    '9999999999',
                    '12345678909',
                    '918328328322',
                    '33459830234',
                    '8234592345',
                    '23439433343',
                ],
                not_phone_nums=[
                    '99999',
                    '12349',
                    '918328328322',
                    '33430234',
                    '8234592345',
                    '23439433343',
                ],
                int_phone_nums=[
                    9999999999,
                    12345678909,
                    918328328322,
                    33459830234,
                    8234592345,
                    23439433343,
                ],
                float_phone_nums=[
                    9999999999,
                    np.nan,
                    918328328322.0,
                    33459830234,
                    8234592345,
                    np.nan,
                ],
                not_float_phone_nums=[
                    9999999999.00234233,
                    np.nan,
                    918328328322.0,
                    33459830234.234,
                    8234592345,
                    np.nan,
                ],
                bad_column_name_one=[
                    9999999999,
                    np.nan,
                    918328328322.0,
                    33459830234,
                    8234592345,
                    np.nan,
                ],
                bad_column_name_two=[self.fake.phone_number() for _ in range(6)],
            )
        )
        ctypes = infer_column_types(df)
        expected_ctypes = dict(
            phone_num='phone_number',
            str_int_phone_nums='phone_number',
            not_phone_nums='number',
            int_phone_nums='phone_number',
            float_phone_nums='phone_number',
            not_float_phone_nums='number_with_decimals',
            bad_column_name_one='number_with_decimals',
            bad_column_name_two='text',
        )
        self.assertEquals(ctypes, expected_ctypes)

    def test_zip_code_recognition(self):
        df = pd.DataFrame(
            dict(
                zip_code=[self.fake.postcode() for _ in range(6)],
                postal_code=[self.fake.postcode() for _ in range(6)],
                bad_column_name=[self.fake.postcode() for _ in range(6)],
                string_zips=[
                    '12345',
                    '132',
                    '234-3434',
                    '765',
                    '41324-2343',
                    '12342',
                ],
                not_string_postal_zips=[
                    '12',
                    '13234523452',
                    '23-3434',
                    '765,23453',
                    '41324-2343-23423',
                    '12342',
                ],
                int_zips=[
                    12345,
                    53123,
                    323,
                    423,
                    678,
                    67896,
                ],
                not_int_zips=[
                    12345,
                    34553123,
                    323343,
                    423,
                    67834534,
                    67896,
                ],
                bad_column_name_two=[
                    12345,
                    53123,
                    323,
                    423,
                    678,
                    67896,
                ],
            )
        )
        ctypes = infer_column_types(df)
        expected_ctypes = dict(
            zip_code=ZIP_CODE,
            postal_code=ZIP_CODE,
            bad_column_name=NUMBER,
            string_zips=ZIP_CODE,
            not_string_postal_zips=TEXT,
            int_zips=ZIP_CODE,
            not_int_zips=NUMBER,
            bad_column_name_two=NUMBER,
        )
        self.assertEquals(ctypes, expected_ctypes)

    def test_datetime_recognition(self):
        df = pd.DataFrame(
            dict(
                dates=[self.fake.date() for _ in range(6)],
                datetimes=[self.fake.date_time().strftime('%d-%m-%Y %H:%M:%S') for _ in range(6)],
                datetimes_iso=[self.fake.date_time().isoformat() for _ in range(6)],
                dates_w_slash=[self.fake.date_time().strftime('%m/%d/%Y') for _ in range(6)],
                datetimes_w_slash=[
                    self.fake.date_time().strftime('%m/%d/%Y %H:%M:%S') for _ in range(6)
                ],
            )
        )
        print(df)
        ctypes = infer_column_types(df)
        for col in ctypes:
            print(col)
            self.assertEqual(ctypes[col], DATETIME)



import time


class timer(object):
    """
    with timer('metric.metric', tags={ 'key': 'value' }):
        function()
    """
    def __init__(self, metric, tags={}, verbose=True):
        self.metric = metric
        self.start = None
        self.tags = tags
        self.verbose = verbose

    def __enter__(self):
        self.start = time.time()

    def __exit__(self, type, value, traceback):
        # Must convert to milliseconds, see details in
        # https://statsd.readthedocs.io/en/v3.1/timing.html
        dt = int((time.time() - self.start) * 1000)
        if self.verbose:
            print(f'[time] metric: {self.metric}, value: {dt}ms, tags: {self.tags}')



from mage_ai.data_cleaner.statistics.calculator import StatisticsCalculator
from mage_ai.tests.base_test import TestCase
import pandas as pd
import numpy as np


class StatisticsCalculatorTest(TestCase):
    def test_calculate_statistics_overview(self):
        calculator = StatisticsCalculator(
            column_types=dict(
                age='number',
                country='category',
                date_joined='datetime',
                id='number',
                cancelled='true_or_false',
            ),
        )

        df = pd.DataFrame(
            [
                [1, 'JP', '2000-01-01', '1', False],
                [2, 'KO', '2000-12-01', '2', False],
                [1, 'US', '2000-07-01 00:00:00+00:00', '3', False],
                [' ', 'US', '2000-07-01 00:00:00+00:00', '4', False],
                [1, None, '2000-07-01 00:00:00+00:00', '5', False],
                ['', None, '899-07-01 00:00:00+00:00', '6', False],
                [3, 'US', None, '7', True],
            ],
            columns=['age', 'country', 'date_joined', 'id', 'cancelled'],
        )

        data = calculator.calculate_statistics_overview(df, is_clean=False)

        self.assertEqual(data['count'], 7)
        self.assertEqual(data['total_null_value_count'], 6)

        self.assertEqual(data['age/average'], 8 / 5)
        self.assertEqual(data['age/count'], 5)
        self.assertEqual(data['age/count_distinct'], 3)
        self.assertEqual(data['age/max'], 3)
        self.assertEqual(data['age/median'], 1)
        self.assertEqual(data['age/min'], 1)
        self.assertEqual(data['age/sum'], 8)
        self.assertEqual(data['age/null_value_rate'], 2 / 7)
        self.assertEqual(data['age/completeness'], 1 - data['age/null_value_rate'])
        self.assertEqual(data['age/quality'], 'Bad')

        self.assertEqual(data['country/count'], 5)
        self.assertEqual(data['country/count_distinct'], 3)
        self.assertEqual(data['country/mode'], 'US')
        self.assertEqual(data['country/null_value_rate'], 2 / 7)
        self.assertEqual(data['country/completeness'], 1 - data['country/null_value_rate'])
        self.assertEqual(data['country/quality'], 'Bad')

        self.assertEqual(data['date_joined/count'], 5)
        self.assertEqual(data['date_joined/count_distinct'], 3)
        self.assertEqual(data['date_joined/max'], '2000-12-01T00:00:00+00:00')
        self.assertEqual(data['date_joined/median'], '2000-07-01T00:00:00+00:00')
        self.assertEqual(data['date_joined/min'], '2000-01-01T00:00:00+00:00')
        self.assertEqual(data['date_joined/mode'], '2000-07-01T00:00:00+00:00')
        self.assertEqual(data['date_joined/null_value_rate'], 2 / 7)
        self.assertEqual(data['date_joined/completeness'], 1 - data['date_joined/null_value_rate'])
        self.assertEqual(data['date_joined/quality'], 'Bad')

        self.assertEqual(data['id/count'], 7)
        self.assertEqual(data['id/count_distinct'], 7)
        self.assertEqual(data['id/null_value_rate'], 0)
        self.assertEqual(data['id/completeness'], 1)
        self.assertEqual(data['id/quality'], 'Good')

        self.assertEqual(data['cancelled/mode'], False)
        self.assertEqual(data['cancelled/null_value_rate'], 0)
        self.assertEqual(data['cancelled/completeness'], 1)
        self.assertEqual(data['cancelled/quality'], 'Good')

    def test_calculate_statistics_overview_no_cleaning(self):
        calculator = StatisticsCalculator(
            column_types=dict(
                age='number',
                country='category',
                date_joined='datetime',
                id='number',
                cancelled='true_or_false',
            ),
        )

        df = pd.DataFrame(
            [
                [1, 'JP', '2000-01-01', 1, False],
                [2, 'KO', '2000-12-01', 2, False],
                [1, 'US', '2000-07-01 00:00:00+00:00', 3, False],
                [np.nan, 'US', '2000-07-01 00:00:00+00:00', 4, False],
                [1, None, '2000-07-01 00:00:00+00:00', 5, False],
                [np.nan, None, '899-07-01 00:00:00+00:00', 6, False],
                [3, 'US', None, 7, True],
            ],
            columns=['age', 'country', 'date_joined', 'id', 'cancelled'],
        )

        data = calculator.calculate_statistics_overview(df, is_clean=True)

        self.assertEqual(data['count'], 7)
        self.assertEqual(data['total_null_value_count'], 5)

        self.assertEqual(data['age/average'], 8 / 5)
        self.assertEqual(data['age/count'], 5)
        self.assertEqual(data['age/count_distinct'], 3)
        self.assertEqual(data['age/max'], 3)
        self.assertEqual(data['age/median'], 1)
        self.assertEqual(data['age/min'], 1)
        self.assertEqual(data['age/sum'], 8)
        self.assertEqual(data['age/null_value_rate'], 2 / 7)
        self.assertEqual(data['age/completeness'], 1 - data['age/null_value_rate'])
        self.assertEqual(data['age/quality'], 'Bad')

        self.assertEqual(data['country/count'], 5)
        self.assertEqual(data['country/count_distinct'], 3)
        self.assertEqual(data['country/mode'], 'US')
        self.assertEqual(data['country/null_value_rate'], 2 / 7)
        self.assertEqual(data['country/completeness'], 1 - data['country/null_value_rate'])
        self.assertEqual(data['country/quality'], 'Bad')

        self.assertEqual(data['date_joined/count'], 6)
        self.assertEqual(data['date_joined/count_distinct'], 4)
        self.assertEqual(data['date_joined/max'], '2000-12-01T00:00:00+00:00')
        self.assertEqual(data['date_joined/median'], '2000-07-01T00:00:00+00:00')
        self.assertEqual(data['date_joined/min'], '2000-01-01T00:00:00+00:00')
        self.assertEqual(data['date_joined/mode'], '2000-07-01T00:00:00+00:00')
        self.assertEqual(data['date_joined/null_value_rate'], 1 / 7)
        self.assertEqual(data['date_joined/completeness'], 1 - data['date_joined/null_value_rate'])
        self.assertEqual(data['date_joined/quality'], 'Good')

        self.assertEqual(data['id/count'], 7)
        self.assertEqual(data['id/count_distinct'], 7)
        self.assertEqual(data['id/null_value_rate'], 0)
        self.assertEqual(data['id/completeness'], 1)
        self.assertEqual(data['id/quality'], 'Good')

        self.assertEqual(data['cancelled/mode'], False)
        self.assertEqual(data['cancelled/null_value_rate'], 0)
        self.assertEqual(data['cancelled/completeness'], 1)
        self.assertEqual(data['cancelled/quality'], 'Good')

    def test_calculate_statistics_overview_divide_by_zero(self):
        calculator = StatisticsCalculator(
            column_types=dict(age='number', date='datetime', category='category', string='string'),
        )

        df = pd.DataFrame([], columns=['age', 'date', 'category', 'string'])
        data = calculator.calculate_statistics_overview(df, is_clean=False)

        self.assertEqual(data['age/count'], 0)
        self.assertEqual(data['age/count_distinct'], 0)
        self.assertEqual(data['age/null_value_rate'], 0)
        self.assertEqual(data['date/count'], 0)
        self.assertEqual(data['date/count_distinct'], 0)
        self.assertEqual(data['date/null_value_rate'], 0)
        self.assertEqual(data['category/count'], 0)
        self.assertEqual(data['category/count_distinct'], 0)
        self.assertEqual(data['category/null_value_rate'], 0)
        self.assertEqual(data['string/count'], 0)
        self.assertEqual(data['string/count_distinct'], 0)
        self.assertEqual(data['string/null_value_rate'], 0)
        print(data['date/value_counts'])

    def test_calculate_statistics_handle_nulls(self):
        calculator = StatisticsCalculator(
            column_types=dict(
                age='number',
                country='category',
                date_joined='datetime',
                id='number',
                cancelled='true_or_false',
            ),
        )

        df = pd.DataFrame(
            [
                [1, None, None, 1, False],
                [np.nan, None, '2000-12-01', 2, False],
                [np.nan, 'US', '2000-07-01 00:00:00+00:00', 3, False],
                [np.nan, 'US', '2000-07-01 00:00:00+00:00', 4, False],
                [1, None, None, 5, False],
                [np.nan, None, '899-07-01 00:00:00+00:00', 6, False],
                [3, 'JP', None, 7, True],
            ],
            columns=['age', 'country', 'date_joined', 'id', 'cancelled'],
        )

        data = calculator.calculate_statistics_overview(df, is_clean=True)

        self.assertEqual(data['count'], 7)
        self.assertEqual(data['total_null_value_count'], 11)

        self.assertEqual(data['age/average'], 5 / 3)
        self.assertEqual(data['age/count'], 3)
        self.assertEqual(data['age/count_distinct'], 2)
        self.assertEqual(data['age/max'], 3)
        self.assertEqual(data['age/median'], 1)
        self.assertEqual(data['age/min'], 1)
        self.assertEqual(data['age/sum'], 5)
        self.assertEqual(data['age/null_value_rate'], 4 / 7)
        self.assertEqual(data['age/completeness'], 1 - data['age/null_value_rate'])
        self.assertEqual(data['age/quality'], 'Bad')

        self.assertEqual(data['country/count'], 3)
        self.assertEqual(data['country/count_distinct'], 2)
        self.assertEqual(data['country/mode'], 'US')
        self.assertEqual(data['country/null_value_rate'], 4 / 7)
        self.assertEqual(data['country/completeness'], 1 - data['country/null_value_rate'])
        self.assertEqual(data['country/quality'], 'Bad')

        self.assertEqual(data['date_joined/count'], 4)
        self.assertEqual(data['date_joined/count_distinct'], 3)
        self.assertEqual(data['date_joined/max'], '2000-12-01T00:00:00+00:00')
        self.assertEqual(data['date_joined/median'], '2000-07-01T00:00:00+00:00')
        self.assertEqual(data['date_joined/min'], '2000-07-01T00:00:00+00:00')
        self.assertEqual(data['date_joined/mode'], '2000-07-01T00:00:00+00:00')
        self.assertEqual(data['date_joined/null_value_rate'], 3 / 7)
        self.assertEqual(data['date_joined/completeness'], 1 - data['date_joined/null_value_rate'])
        self.assertEqual(data['date_joined/quality'], 'Bad')

        self.assertEqual(data['id/count'], 7)
        self.assertEqual(data['id/count_distinct'], 7)
        self.assertEqual(data['id/null_value_rate'], 0)
        self.assertEqual(data['id/completeness'], 1)
        self.assertEqual(data['id/quality'], 'Good')

        self.assertEqual(data['cancelled/mode'], False)
        self.assertEqual(data['cancelled/null_value_rate'], 0)
        self.assertEqual(data['cancelled/completeness'], 1)
        self.assertEqual(data['cancelled/quality'], 'Good')



from mage_ai.data_cleaner.cleaning_rules.base import BaseRule
from mage_ai.data_cleaner.column_type_detector import NUMBER_TYPES
from mage_ai.data_cleaner.transformer_actions.constants import ActionType, Axis
import numpy as np


class RemoveCollinearColumns(BaseRule):
    EPSILON = 1e-12
    MIN_ENTRIES = 3
    ROW_SAMPLE_SIZE = 300
    VIF_UB = 5

    def __init__(self, df, column_types, statistics):
        super().__init__(df, column_types, statistics)
        self.numeric_df, self.numeric_columns = self.filter_numeric_types()
        self.numeric_indices = np.arange(len(self.numeric_df))

    def evaluate(self):
        suggestions = []
        if self.numeric_df.empty or len(self.numeric_df) < self.MIN_ENTRIES:
            return suggestions
        collinear_columns = []
        self.numeric_df['intercept'] = np.ones(len(self.numeric_df))
        for column in self.numeric_columns[:-1]:
            variance_inflation_factor = self.get_variance_inflation_factor(column)
            if variance_inflation_factor > self.VIF_UB:
                collinear_columns.append(column)
                self.numeric_df.drop(column, axis=1, inplace=True)
        if len(collinear_columns) != len(self.numeric_columns) - 1:
            # check the final column if and only if there are other columns to compare it to
            column = self.numeric_columns[-1]
            variance_inflation_factor = self.get_variance_inflation_factor(column)
            if variance_inflation_factor > self.VIF_UB:
                collinear_columns.append(column)
        if len(collinear_columns) != 0:
            suggestions.append(
                self._build_transformer_action_suggestion(
                    'Remove collinear columns',
                    'Delete these columns to remove redundant data and increase data quality.',
                    ActionType.REMOVE,
                    action_arguments=collinear_columns,
                    axis=Axis.COLUMN,
                )
            )
        return suggestions

    def filter_numeric_types(self):
        numeric_columns = []
        numeric_df = self.df.copy()
        for column in self.df_columns:
            if self.column_types[column] in NUMBER_TYPES:
                numeric_df.loc[:, column] = numeric_df.loc[:, column].astype(float)
                numeric_columns.append(column)
            else:
                numeric_df.drop(column, axis=1, inplace=True)
        numeric_df = numeric_df.dropna(axis=0)
        return numeric_df, numeric_columns

    def get_variance_inflation_factor(self, column):
        """
        Variance Inflation Factor = 1 / (1 - <coefficient of determination on column k>)
        Measures increase in regression model variance due to collinearity
        => column k is multicollinear with others if model predicting its value
        has this variance inflation greater than some amount
        """
        if self.numeric_df.empty:
            raise RuntimeError('No other columns to compare \'{column}\' against')
        if len(self.numeric_df) > self.ROW_SAMPLE_SIZE:
            sample = self.numeric_df.sample(self.ROW_SAMPLE_SIZE)
        else:
            sample = self.numeric_df

        responses = sample[column].to_numpy()
        predictors = sample.drop(column, axis=1).to_numpy()
        params, _, _, _ = np.linalg.lstsq(predictors, responses, rcond=None)

        mean = responses.mean()
        centered_predictions = predictors @ params - mean
        sum_sq_model = np.sum(centered_predictions * centered_predictions)
        centered_responses = responses - mean
        sum_sq_to = np.sum(centered_responses * centered_responses)
        r_sq = sum_sq_model / sum_sq_to if sum_sq_to else 0
        return 1 / (1 - r_sq + self.EPSILON)



STATUS_NOT_APPLIED = 'not_applied'
STATUS_COMPLETED = 'completed'

class BaseRule:
    def __init__(self, df, column_types, statistics):
        self.df = df
        self.df_columns = df.columns.tolist()
        self.column_types = column_types
        self.statistics = statistics

    def evaluate(self):
        """Evaluate data cleaning rule and generate suggested actions
        Returns
        -------
        A list of suggested actions
        """
        return []

    def _build_transformer_action_suggestion(
        self,
        title,
        message,
        action_type,
        action_arguments=[],
        action_code='',
        action_options={},
        action_variables={},
        axis='column',
        outputs=[],
    ):
        return dict(
            title=title,
            message=message,
            status=STATUS_NOT_APPLIED,
            action_payload=dict(
                action_type=action_type,
                action_arguments=action_arguments,
                action_code=action_code,
                action_options=action_options,
                action_variables=action_variables,
                axis=axis,
                outputs=outputs,
            ),
        )



from mage_ai.data_cleaner.column_type_detector import (
    NUMBER,
    NUMBER_WITH_DECIMALS,
    DATETIME,
    PHONE_NUMBER,
    ZIP_CODE,
)
from mage_ai.data_cleaner.transformer_actions.constants import CURRENCY_SYMBOLS
import pandas as pd
import numpy as np

COLUMN_NAME_QUOTE_CHARS = '+=-*&^%$! ?~|<>(){}[],.'


def clean_series(series, column_type, dropna=True):
    series_cleaned = series.apply(lambda x: x.strip(' \'\"') if type(x) is str else x)
    series_cleaned = series_cleaned.map(
        lambda x: x if (not isinstance(x, str) or (len(x) > 0 and not x.isspace())) else np.nan
    )
    if dropna:
        series_cleaned = series_cleaned.dropna()

    if series_cleaned.count() == 0:
        return series_cleaned

    dtype = type(series_cleaned.dropna().iloc[0])
    if column_type == NUMBER or column_type == NUMBER_WITH_DECIMALS:
        is_percent = False
        if dtype is str:
            series_cleaned = series_cleaned.str.replace(',', '')
            if series_cleaned.str.count(CURRENCY_SYMBOLS).sum() != 0:
                series_cleaned = series_cleaned.str.replace(CURRENCY_SYMBOLS, '')
            elif series_cleaned.str.contains('%').sum() != 0:
                is_percent = True
                series_cleaned = series_cleaned.str.replace('%', '')
            series_cleaned = series_cleaned.str.replace(' ', '')
        if column_type == NUMBER:
            try:
                series_cleaned = series_cleaned.astype(int)
            except ValueError:
                series_cleaned = series_cleaned.astype(float)
        else:
            series_cleaned = series_cleaned.astype(float)
        if is_percent:
            series_cleaned /= 100
    elif column_type == DATETIME:
        series_cleaned = pd.to_datetime(series_cleaned, errors='coerce', infer_datetime_format=True)
    elif column_type == PHONE_NUMBER and dtype is not str:
        series_cleaned = series_cleaned.astype(str)
        series_cleaned = series_cleaned.str.replace(r'\.\d*', '')
    elif column_type == ZIP_CODE and dtype is not str:
        series_cleaned = series_cleaned.astype(str)
    return series_cleaned


def clean_dataframe(df, column_types, dropna=True):
    return df.apply(lambda col: clean_series(col, column_types[col.name], dropna=dropna), axis=0)


def is_numeric_dtype(df, column, column_type):
    return column_type in [NUMBER, NUMBER_WITH_DECIMALS] or issubclass(
        df[column].dtype.type, np.number
    )


def wrap_column_name(name: str) -> str:
    if any(symbol in name for symbol in COLUMN_NAME_QUOTE_CHARS):
        name = f'"{name}"'
    return name



from IPython import get_ipython
from IPython.display import IFrame, Javascript, display
from mage_ai.server.app import (
    clean_df,
    clean_df_with_pipeline,
    connect_df,
    kill as kill_flask,
    launch as launch_flask,
)
from mage_ai.server.constants import SERVER_PORT

IFRAME_HEIGHT = 1000
MAX_NUM_OF_ROWS = 100_000


def launch():
    thread = launch_flask()
    display_inline_iframe()
    return thread


def kill():
    kill_flask()


def display_inline_iframe():
    path_to_local_server = f'http://localhost:{SERVER_PORT}'
    if type(get_ipython()).__module__.startswith('google.colab'):
        display(Javascript("""
            (async ()=>{
                fm = document.createElement('iframe')
                fm.src = await google.colab.kernel.proxyPort(%s)
                fm.width = '95%%'
                fm.height = '%d'
                fm.frameBorder = 0
                document.body.append(fm)
            })();
            """ % (SERVER_PORT, IFRAME_HEIGHT)))
    else:
        display(IFrame(path_to_local_server, width='95%', height=1000))


def connect_data(df, name):
    if df.shape[0] > MAX_NUM_OF_ROWS:
        feature_set, _ = connect_df(df.sample(MAX_NUM_OF_ROWS), name)
    else:
        feature_set, _ = connect_df(df, name)
    return feature_set


def clean(df, pipeline_uuid=None, pipeline_path=None):
    if pipeline_uuid is not None:
        df_clean = clean_df_with_pipeline(df, id=pipeline_uuid)
    elif pipeline_path is not None:
        df_clean = clean_df_with_pipeline(df, path=pipeline_path)
    else:
        _, df_clean = clean_df(df)
    return df_clean


def init(api_key):
    # verify api_key with Mage backend
    pass



from mage_ai.data_cleaner.cleaning_rules.base import BaseRule
from mage_ai.data_cleaner.shared.utils import wrap_column_name
from mage_ai.data_cleaner.transformer_actions.constants import (
    ActionType,
    Axis,
)

REMOVE_OUTLIERS_TITLE = 'Remove outliers'


class RemoveOutliers(BaseRule):
    def evaluate(self):
        suggestions = []
        for c in self.df_columns:
            if f'{c}/outlier_count' not in self.statistics:
                continue
            outlier_count = self.statistics[f'{c}/outlier_count']
            if outlier_count == 0:
                continue
            std = self.statistics[f'{c}/std']
            avg = self.statistics[f'{c}/average']
            upper = avg + 3 * std
            lower = avg - 3 * std

            wrapped_c = wrap_column_name(c)
            suggestions.append(
                self._build_transformer_action_suggestion(
                    REMOVE_OUTLIERS_TITLE,
                    f'Remove {outlier_count} outlier(s) to reduce the amount of noise in the data.',
                    ActionType.FILTER,
                    action_arguments=[c],
                    action_code=f'{wrapped_c} <= {upper} and {wrapped_c} >= {lower}',
                    axis=Axis.ROW,
                )
            )
        return suggestions



import setuptools


def readme():
    with open("README.md", encoding="utf8") as f:
        README = f.read()
    return README


requirements, dependency_links = [], []
with open('requirements.txt') as f:
    for line in f.read().splitlines():
        if line.startswith('-e git+'):
            dependency_links.append(line.replace('-e ', ''))
        else:
            requirements.append(line)

setuptools.setup(
    name='mage-ai',
    version='0.0.3',
    author='Mage',
    author_email="sales@mage.ai",
    description='Mage - An open-source data management platform that helps you '
                'clean data and prepare it for training AI/ML models',
    long_description=readme(),
    long_description_content_type="text/markdown",
    url='https://github.com/mage-ai/mage-ai',
    packages=setuptools.find_packages('./'),
    include_package_data=True,
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: OS Independent',
    ],
    install_requires=requirements,
    dependency_links=dependency_links,
    python_requires='>=3.6',
    entry_points={},
    extras_require={}
)



from mage_ai.data_cleaner.cleaning_rules.remove_outliers import RemoveOutliers
from mage_ai.tests.base_test import TestCase
import pandas as pd


class RemoveOutliersTests(TestCase):
    def test_evaluate(self):
        df = pd.DataFrame(
            [
                [1, 1, 1],
                [1, 2, 997],
                [1, 3, 998],
                [1, 4, 999],
                [1, 5, 1000],
                [1, 6, 997],
                [1, 7, 998],
                [1, 8, 999],
                [1, 9, 1000],
                [1, 10, 997],
                [1, 11, 998],
                [1, 12, 999],
            ],
            columns=['number1', 'number2', 'number3'],
        )
        column_types = {
            'number1': 'number',
            'number2': 'number',
            'number3': 'number',
        }
        statistics = {
            'number1/average': 1,
            'number1/outlier_count': 0,
            'number1/std': 0,
            'number2/average': 6.5,
            'number2/outlier_count': 0,
            'number2/std': 3.6,
            'number3/average': 915.25,
            'number3/outlier_count': 1,
            'number3/std': 287.92,
        }
        result = RemoveOutliers(df, column_types, statistics).evaluate()

        self.assertEqual(
            result,
            [
                dict(
                    title='Remove outliers',
                    message='Remove 1 outlier(s) to reduce the amount of noise in the data.',
                    status='not_applied',
                    action_payload=dict(
                        action_type='filter',
                        action_arguments=['number3'],
                        action_code='number3 <= 1779.01 and number3 >= 51.49000000000001',
                        action_options={},
                        action_variables={},
                        axis='row',
                        outputs=[],
                    ),
                )
            ],
        )

    def test_evaluate_wrapped_column_name(self):
        df = pd.DataFrame(
            [
                [1, 1, 1],
                [1, 2, 997],
                [1, 3, 998],
                [1, 4, 999],
                [1, 5, 1000],
                [1, 6, 997],
                [1, 7, 998],
                [1, 8, 999],
                [1, 9, 1000],
                [1, 10, 997],
                [1, 11, 998],
                [1, 12, 999],
            ],
            columns=['number1.', 'num,ber2', 'number ([3])'],
        )
        column_types = {
            'number1.': 'number',
            'num,ber2': 'number',
            'number ([3])': 'number',
        }
        statistics = {
            'number1./average': 1,
            'number1./outlier_count': 0,
            'number1./std': 0,
            'num,ber2/average': 6.5,
            'num,ber2/outlier_count': 0,
            'num,ber2/std': 3.6,
            'number ([3])/average': 915.25,
            'number ([3])/outlier_count': 1,
            'number ([3])/std': 287.92,
        }
        result = RemoveOutliers(df, column_types, statistics).evaluate()

        self.assertEqual(
            result,
            [
                dict(
                    title='Remove outliers',
                    message='Remove 1 outlier(s) to reduce the amount of noise in the data.',
                    status='not_applied',
                    action_payload=dict(
                        action_type='filter',
                        action_arguments=['number ([3])'],
                        action_code='"number ([3])" <= 1779.01 and "number ([3])" >= 51.49000000000001',
                        action_options={},
                        action_variables={},
                        axis='row',
                        outputs=[],
                    ),
                )
            ],
        )
